/*
 * Main miner class
 * Author: Michael Mampaey
 * Date: 2009-06-13
 *
 */

#include <iostream>
#include <stdio.h>
#include <time.h>
#include <string.h>

#include "aux.h"
#include "miner.h"

using namespace std;



const string Miner::name        = "MuMiner";
const string Miner::version     = "v1.6.20090613";
const string Miner::infostring  = Miner::name + " " + Miner::version + " -- http://www.adrem.ua.ac.be";
const string Miner::author      = "Michael Mampaey";



Miner::Miner() : data(0), out(0), to_std_out(0), pattern_type(RULES), redundancy_pruning(MINIMAL),
generators(false), ndi(false), maxk(-1), header(false), verbose(false), sup_ent_trie(0), resorted(0)
{
	infilename  = "";
	outfilename = "";
}


Miner::~Miner()
{
}


void Miner::initiate()
{
	data = new Data((char*)(infilename.c_str()));
	if (outfilename.size()) {
		out = fopen(outfilename.c_str(), "wt");
	}
	if (header) {
		attribute_names.clear();
	}

	sup_ent_trie = new Trie<pid>(pid(0,0));
	partitions = new int*[32]; // assume attribute sets of size < 32
	checkers = new unsigned*[32];
	for (int i=0; i<32; ++i) {
		partitions[i] = 0;
		checkers[i] = 0;
	}
	resorted = 0;

	cand_cnt = 0;
	set_cnt  = 0;
	rule_cnt = 0;

	return;
}


void Miner::cleanup()
{
	delete data;
	data = 0;

	if (out) {
		fclose(out);
		out = 0;
	}

	attribute_names.clear();

	delete sup_ent_trie;
	sup_ent_trie = 0;
	for (int i=0; i<32; ++i) {
		if (partitions[i])
			delete [] partitions[i];
		if (checkers[i]) {
			delete [] checkers[i];
		}
	}
	delete [] partitions;
	partitions = 0;
	delete [] checkers;
	checkers = 0;
	delete [] resorted;
	resorted = 0;

	return;
}


void Miner::mine()
{
	clock_t read_t=0, mine_t=0, end_t=0;

	initiate();

	if (verbose) {
		cout << infostring << endl << endl;
		cout << "Progress:" << endl;
		cout << '\t' << "Reading data file ...\t"<<flush;
		read_t = clock();
	}

	set<Item> *root = read_data();

	if (verbose){
		cout << "Done. (" << t_cnt << " transactions)" << endl;
		cout << '\t' << "Pruning attributes ...\t" << flush;
		mine_t = clock();
	}

	root = resort_attributes(root);

	pair<Item**,int> allattributes = prune_attributes(root);
	delete root;

	if (verbose) {
		cout << "Done. (" << allattributes.second << " attributes)" << endl;
		cout << '\t' << "Running " << name << " ...\t" << flush;
		if (to_std_out) cout << endl << endl << " *** PATTERNS *** " << endl << endl;
	}

	if (maxk==-1)
		maxk = allattributes.second;
	int *attrset = new int[maxk];
	int *closure = 0;
	int clsize = 0;
	if (pattern_type==SETS)
		closure = new int[maxk];
	Node<pid> *rel_node = sup_ent_trie->add(0, 0, pid(t_cnt, 0.0) );
	set<Item> delzero;
	for (int i=0; i<allattributes.second; ++i) {
		if (allattributes.first[i]->entropy==0.0) {
			if (pattern_type==SETS) {
				if (!generators)
					closure[clsize++] = allattributes.first[i]->id;
			}
			else if (pattern_type==RULES) {
				print_rule(0, 0, allattributes.first[i]->id, 0.0, 1.0);
			}
			delete allattributes.first[i];
			allattributes.first[i] = 0;
		}
	}
	int i = 0, d = 0;
	while (i + d < allattributes.second) {
		if (allattributes.first[i+d]==0)
			++d;
		else {
			if (d)
				allattributes.first[i] = allattributes.first[i+d];
			++i;
		}
	}
	allattributes.second -= d;

	if (maxk!=0) {
		set_cnt += (1 << clsize) - 1;
		set_mine(allattributes.first, allattributes.second, attrset, 1, rel_node, closure, clsize);
	}

	delete [] attrset;
	if (pattern_type==SETS)
		delete [] closure;
	for (int i=0; i<allattributes.second; ++i)
		delete allattributes.first[i];
	delete [] allattributes.first;

	if (verbose) {
		end_t = clock();
		if (to_std_out) cout << endl << " *** " << endl << endl;
		else cout << "Done." << endl;
		cout << "Results:" << endl;
		cout << '\t' << cand_cnt << '\t' << "Candidate attribute sets" << endl;
		cout << '\t' << set_cnt  << '\t' << "Low entropy attribute sets" << endl;
		if (pattern_type==RULES) {
			if (redundancy_pruning==NONE)
				cout << '\t' << rule_cnt << '\t' << "Association rules" << endl;
			else {
				cout << '\t' << rule_cnt << '\t';
				cout << ( redundancy_pruning==MINIMAL ? "Minimal" : "Non-redundant" ) ;
				cout << " association rules" << endl;
			}

		}
		cout << "Time:" << endl;
		cout << '\t' << "Read"  << '\t' << (mine_t - read_t)/double(CLOCKS_PER_SEC) << " sec" << endl;
		cout << '\t' << "Mine"  << '\t' << (end_t - mine_t)/double(CLOCKS_PER_SEC)  << " sec" << endl;
		cout << '\t' << "Total" << '\t' << (end_t - read_t)/double(CLOCKS_PER_SEC)  << " sec" << endl;
	}

	cleanup();

	return;
}


set<Item> *Miner::read_data()
{
	t_cnt = 0;
	set<Item> *root = new set<Item>();

	if (header) {
		Transaction<string> *h = data->getHeader();
		for (int i=0; i<h->length; ++i)
			attribute_names.push_back(h->t[i]);
		delete h;
	}

	set<Item>::iterator it;
	while (Transaction<int> * t = data->getNext()) {
		for (int i=0; i<t->length; ++i) {
			Item tmp_attr = Item(t->t[i], t->t[i], t->t[i]);
			it = root->find(tmp_attr);
			if (it == root->end())
				it = root->insert(tmp_attr).first;
			it->tidlist->add(t_cnt);
		}
		delete t;
		++t_cnt;
	}

	if (header) {
		for (unsigned i=attribute_names.size(); i<root->size(); ++i)
			attribute_names.push_back(string("?"));
	}

	return root;
}


set<Item> * Miner::resort_attributes(set<Item> *root) {

	set<Item> tmproot;
	set<Item>::iterator it;

	while ((it=root->begin()) != root->end()) {
		unsigned isup = it->tidlist->size;
		double ient = (isup==0 || isup==t_cnt) ? 0.0 : h(isup, t_cnt);
		Item attr(*it);
		attr.entropy = ient;
		attr.support = isup;
		tmproot.insert(attr);
		root->erase(*it);
	}
	delete root;

	set<Item> *newroot = new set<Item>();
	resorted = new int[tmproot.size()+1];
	int cnt = 1;

	for(it=tmproot.begin(); it!=tmproot.end(); ++it, ++cnt) {
		resorted[cnt] = it->id;
		Item attr(*it);
		attr.id = cnt;
		newroot->insert(attr);
	}

	return newroot;
}


pair<Item**,int> Miner::prune_attributes(set<Item> *root)
{
	int size = 0;
	Item **allattributes = new Item*[root->size()];

	set<Item>::reverse_iterator it;
	while ((it=root->rbegin()) != root->rend()) { // order
		++cand_cnt;
		unsigned isup = it->tidlist->size;
		double ifreq = isup / (double)(t_cnt);
		if (ifreq >= minsup && ifreq <= (1.0 - minsup)) {
			double ient = (isup==0 || isup==t_cnt) ? 0.0 : h(isup, t_cnt);
			if (ient <= maxentropy) {
				Item *attr = new Item(*it);
				attr->support = isup;
				attr->entropy = ient;
				sup_ent_trie->add((int*)&(it->id), 1, pid(isup, ient));
				attr_ents[it->id] = ient;
				allattributes[size++] = attr;
			}
		}
		root->erase(*it);
	}

	return pair<Item**,int>(allattributes, size);
}


void Miner::set_mine(Item **attributes, int alen, int *attrset, int depth, Node<pid>* rel_node, int *closure, int close_len)
{
	if (depth >= maxk) {
		for (int it1=0; it1<alen; ++it1) {
			const Item &attr1 = *(attributes[it1]);
			attrset[depth-1] = attr1.id;
			if (pattern_type==RULES) {
				set_cnt += 1;
				rule_mine(attrset, depth, attr1.entropy, 0, depth, 0, 0);
			}
			else {
				set_cnt += (1 << close_len);
				print_set(attrset, depth, closure, close_len, attr1.entropy);
			}
		}
		return;
	}

	for (int it1=0; it1<alen; ++it1) {
		const Item &attr1 = *(attributes[it1]);
		int close_len2 = close_len;
		attrset[depth-1] = attr1.id;
		Item **children = new Item*[it1];
		int children_size = 0;

		Node<pid> *rel_node2 = sup_ent_trie->locate(attrset[depth-1], rel_node);

		for (int it2=0; it2<it1; ++it2) {
			const Item &attr2 = *(attributes[it2]);
			Item new_attr(attr2.id, false);
			++cand_cnt;
			attrset[depth] = attr2.id;

			if (depth < 2) {
				new_attr.set_tid_list(attr1.tidlist->intersect(attr2.tidlist));
				new_attr.support = new_attr.tidlist->size;
			}
			else {
				if (depth == 2) new_attr.set_tid_list(attr1.tidlist->difference(attr2.tidlist));
				else new_attr.set_tid_list(attr2.tidlist->difference(attr1.tidlist));
				new_attr.support = attr1.support - new_attr.tidlist->size;
			}
			sup_ent_trie->add(attrset+depth, 1, pid(new_attr.support, -1.0), rel_node2);

			int *partition = qie(attrset, depth+1);
			double ent = 0.0;
			bool ndiborder = false;
			if (partition) {
				ent = h(partition, 1<<(depth+1), t_cnt);
				if (ndi)
					ndiborder = inarray(partition, 1<<(depth+1), 0);
			}
			else {
				sup_ent_trie->delete_leaf(attrset+depth, 1, rel_node2);
				continue;
			}
			new_attr.entropy = ent;
			sup_ent_trie->add(attrset+depth, 1, pid(new_attr.support,new_attr.entropy), rel_node2); //overwrites

			bool eq1 = same(new_attr.entropy, attr1.entropy);
			bool eq2 = same(new_attr.entropy, attr2.entropy);

			if ( (pattern_type==RULES && redundancy_pruning>NONE && (eq1 || eq2)) || (pattern_type==SETS && eq1) ) {
				if (pattern_type==SETS) {
					if (!generators)
						closure[close_len2++] = attr2.id;
				}
				else {
					++set_cnt;
					if (eq1) {
						double rhsent = attr_ents[attr2.id];
						if (!is_antecedent_redundant(attrset, depth, attr2.id, attrset, depth+1, rhsent, 1.0)) {
							++rule_cnt;
							print_rule(attrset, depth, attr2.id, attr1.entropy, 1.0);
						}
					}
					if (eq2) {
						double rhsent = attr_ents[attr1.id];
						int *sub = make_subset(attrset, depth+1, depth-1);
						if (!is_antecedent_redundant(sub, depth, attr1.id, attrset, depth+1, rhsent, 1.0)) {
							++rule_cnt;
							print_rule(sub, depth, attr1.id, attr2.entropy, 1.0);
						}
						delete [] sub;
					}
					for (int kk=0; kk<depth-1; ++kk) { // check if generator otherwise
						double subent = sub_h(partition, 1<<(depth+1), t_cnt, kk);
						if (same(subent, new_attr.entropy)) {
							double rhsent = attr_ents[attrset[kk]];
							int *sub = make_subset(attrset, depth+1, kk);
							if (!is_antecedent_redundant(sub, depth, attrset[kk], attrset, depth+1, rhsent, 1.0)) {
								++rule_cnt;
								print_rule(sub, depth, attrset[kk], new_attr.entropy, 1.0);
							}
							delete [] sub;
						}
					}

				}
			}
			else if (new_attr.entropy <= maxentropy) {
				bool insert_attr = true;
				if (pattern_type==RULES && redundancy_pruning>NONE) {
					for (int kk=0; kk<depth-1; ++kk) { // check if generator after all
						double subent = sub_h(partition, 1<<(depth+1), t_cnt, kk);
						if (same(subent, new_attr.entropy)) {
							double rhsent = attr_ents[attrset[kk]];
							int *sub = make_subset(attrset, depth+1, kk);
							if (!is_antecedent_redundant(sub, depth, attrset[kk], attrset, depth+1, rhsent, 1.0)) {
								++rule_cnt;
								print_rule(sub, depth, attrset[kk], new_attr.entropy, 1.0);
							}
							delete [] sub;
							insert_attr = false;
						}
					}
				}
				else if (pattern_type==SETS && generators) {
					for (int kk=0; kk<depth; ++kk) { // check if generator after all
						double subent = sub_h(partition, 1<<(depth+1), t_cnt, kk);
						if (same(subent, new_attr.entropy)) {
							insert_attr = false;
							break;
						}
					}
				}
				if (redundancy_pruning==NONE || insert_attr) {
					if (!(ndi && ndiborder)) {
						Item *the_new_attr = new Item(new_attr);
						children[children_size++] = the_new_attr;
					}
				}
			}
			else {
				sup_ent_trie->delete_leaf(attrset+depth, 1, rel_node2);
			}
		} // it2 loop

		if (pattern_type==RULES) {
			set_cnt += 1;
			rule_mine(attrset, depth, attr1.entropy, 0, depth, 0, 0);
		}
		else {
			set_cnt += (1 << close_len2);
			print_set(attrset, depth, closure, close_len2, attr1.entropy);
		}

		set_mine(children, children_size, attrset, depth+1, rel_node2, closure, close_len2);
		for (int i=0; i<children_size; ++i)
			delete children[i];
		delete [] children;

	} // it1 loop

	return;
}


int *Miner::get_partition(int size) const
{
	if (partitions[size]==0) {
		partitions[size] = new int[1 << size];
	}
	return partitions[size];
}


int *Miner::qie(const int *attrset, int len) const
{
	int *partition = get_partition(len);
	sup_ent_trie->get_all_values(attrset, len, (void*&)partition);
	if (partition==0) return 0;

	unsigned plen = (1 << len);
	for (int j=0; j<len; ++j) {
		unsigned ch = ((unsigned)1<<j);
		for (unsigned i=0; i<plen; ++i) {
			if ( i & ch )
				partition[i-ch] -= partition[i];
		}
	}

	return partition;
}


unsigned *Miner::get_checker(int size) const
{
	if (checkers[size]==0) {
		checkers[size] = new unsigned[1<<size];
	}
	return checkers[size];
}


void Miner::rule_mine(int *full, int len, double entropy, int *lhs, int llen, int *rhs, int rlen, int q, unsigned *checker, unsigned indx)
{
	// assume lhs ordered < (?)

	if (llen<=1)
		return;
	if (rlen==0) {
		lhs = full;
		rhs = new int[len];
		checker = get_checker(len);
		for (int i=(1<<len)-1; i>0; --i)
			checker[i] = 1;
		checker[0] = 0;
	}

	double mu;

	Node<pid> *rpos = sup_ent_trie->locate(rhs, rlen);
	if (rpos==0)
		return;

	unsigned bitmask = (1<<len) - 1;
	int *newlhs = new int[llen-1];
	for (int i=len-1; i>=q; --i) {
		rhs[rlen] = full[i];
		unsigned indx2 = indx | (1 << i);
		bool test = checker[indx];
		int j = 0;
		while (test && j<q) {
			test &= checker[indx2 & (bitmask-(1<<j))];
			++j;
		}
		if (test) {
			if (llen-1) {
				int ii = 0;
				while (lhs[ii]!=full[i]) ++ii;
				newlhs = make_subset_into(lhs, llen, ii, newlhs);
			}
		}
		else {
			double lhsent = sup_ent_trie->find(full, len, rhs, rlen+1).second; //
			double rhsent = sup_ent_trie->find(rhs[rlen], rpos).second; //
			if (lhsent == -1 || rhsent == -1) {
				continue;
			}

			mu = (lhsent + rhsent - entropy) / rhsent;

			if (llen-1 || mu>=minmu) {
				int ii = 0;
				while (lhs[ii]!=full[i]) ++ii;
				newlhs = make_subset_into(lhs, llen, ii, newlhs);
			}

			if (mu >= minmu) {
				checker[indx2] = 0;
				if (redundancy_pruning<LOSSY || !is_consequent_redundant(full, len, rhs, rlen+1, lhsent, mu)) {
					if (redundancy_pruning<LOSSY || !is_antecedent_redundant(newlhs, llen-1, rhs, rlen+1, full, len, rhsent, mu)) {
						++rule_cnt;
						print_rule(newlhs, llen-1, rhs, rlen+1, entropy, mu);
					}
				}
			}
		}

		if (llen-1)
			rule_mine(full, len, entropy, newlhs, llen-1, rhs, rlen+1, i+1, checker, indx2);

	}
	delete [] newlhs;

	if (rlen==0) {
		delete [] rhs;
	}

	return;
}


bool Miner::is_consequent_redundant(const int *full, int len, const int *rhs, int rlen, double lhsent, double mu) const
{
	if (rlen<=1)
		return false;

	for (int i=0; i<rlen; ++i) {
		double rhsent = sup_ent_trie->find(rhs, rlen, i).second; // ?
		double fullent = sup_ent_trie->find(full, len, i).second;
		double newmu = (lhsent + rhsent - fullent) / rhsent;
		if (mu < newmu) {
			return true;
		}
	}

	return false;
}


bool Miner::is_antecedent_redundant(const int *lhs, int llen, int rhs, const int *full, int len, double rhsent, double mu) const
{
	if (llen<=1)
		return false;

	int *newfull2 = new int[2];

	for (int i=0; i<llen; ++i) {
		newfull2[0] = rhs;
		newfull2[1] = lhs[i];
		sort(newfull2, newfull2+2);
		double lhsent1, lhsent2, fullent1, fullent2, mu1, mu2;
		lhsent1 = sup_ent_trie->find(lhs, llen, i).second;
		lhsent2 = attr_ents[lhs[i]];
		fullent1 = sup_ent_trie->find(full, llen+1, lhs+i, 1).second;
		fullent2 = sup_ent_trie->find(newfull2, 2).second;
		mu1 = (lhsent1 + rhsent - fullent1) / rhsent;
		mu2 = (lhsent2 + rhsent - fullent2) / rhsent;
		if (mu <= mu1 + mu2) {
			delete [] newfull2;
			return true;
		}
	}

	delete [] newfull2;

	return false;
}


bool Miner::is_antecedent_redundant(const int *lhs, int llen, const int *rhs, int rlen, const int *full, int len, double rhsent, double mu) const
{
	if (llen <=1)
		return false;

	int *newfull2 = new int[rlen+1];

	for (int i=0; i<llen; ++i) {
		memcpy(newfull2, rhs, sizeof(int)*rlen);
		newfull2[rlen] = lhs[i];
		sort(newfull2, newfull2+rlen+1);
		double lhsent1, lhsent2, fullent1, fullent2, mu1, mu2;
		lhsent1 = sup_ent_trie->find(lhs, llen, i).second;
		lhsent2 = attr_ents[lhs[i]];
		fullent1 = sup_ent_trie->find(full, len, lhs+i, 1).second;
		fullent2 = sup_ent_trie->find(newfull2, rlen+1).second;
		mu1 = (lhsent1 + rhsent - fullent1) / rhsent;
		mu2 = (lhsent2 + rhsent - fullent2) / rhsent;
		if (mu <= mu1 + mu2) {
			delete [] newfull2;
			return true;
		}
	}

	delete [] newfull2;
	return false;
}


void Miner::print_set(const int *aset, int len, const int *closure, int close_len, double entropy, int d) const
{
	if (!(out || to_std_out)) return;

	int *aset2;
	if (close_len || sort_output) {
		aset2 = new int[len+close_len];
		memcpy(aset2, aset, sizeof(int)*len);
		memcpy(aset2+len, closure, sizeof(int)*close_len);
		if (sort_output)
			sort(aset2, aset2+len+close_len);
	}
	else {
		aset2 = (int*)aset;
	}

	if (out) {
		if (header)
			for (int k=0; k<len+close_len; ++k) fprintf(out, "%s ", attribute_names[resorted[aset2[k]]].c_str());
		else
			for (int k=0; k<len+close_len; ++k) fprintf(out, "%d ", resorted[aset2[k]]);
		fprintf(out, "(%f)\n", entropy);
	}
	else if (to_std_out) {
		if (header)
			for (int k=0; k<len+close_len; ++k) printf("%s ", attribute_names[resorted[aset2[k]]].c_str());
		else
			for (int k=0; k<len+close_len; ++k) printf("%d ", resorted[aset2[k]]);
		printf("(%f)\n", entropy);
	}

	if (close_len || sort_output)
		delete [] aset2;
	for (int c=d; c<close_len; ++c) {
		int *cl2 = make_subset(closure, close_len, c);
		print_set(aset, len, cl2, close_len-1, entropy, c);
		delete [] cl2;
	}

	return;
}


void Miner::print_rule(const int *lhs, int llen, const int *rhs, int rlen, double entropy, double mu) const
{
	if (!(out||to_std_out)) return;

	int *lhs2, *rhs2;
	if (sort_output) {
		lhs2 = new int[llen];
		rhs2 = new int[rlen];
		memcpy(lhs2, lhs, sizeof(int)*llen);
		memcpy(rhs2, rhs, sizeof(int)*rlen);
		sort(lhs2, lhs2+llen);
		sort(rhs2, rhs2+rlen);
	}
	else {
		lhs2 = (int*)lhs;
		rhs2 = (int*)rhs;
	}

	if (out) {
		if (header)
			for (int k=0; k< llen; ++k) fprintf(out, "%s ", attribute_names[resorted[lhs2[k]]].c_str());
		else
			for (int k=0; k< llen; ++k) fprintf(out, "%d ", resorted[lhs2[k]]);
		fprintf(out, "=> ");
		if (header)
			for (int k=0; k< rlen; ++k) fprintf(out, "%s ", attribute_names[resorted[rhs2[k]]].c_str());
		else
			for (int k=0; k< rlen; ++k) fprintf(out, "%d ", resorted[rhs2[k]]);
		fprintf(out, "(%f, %f)\n", entropy, mu);
	}
	else if (to_std_out) {
		if (header)
			for (int k=0; k< llen; ++k) printf("%s ", attribute_names[resorted[lhs2[k]]].c_str());
		else
			for (int k=0; k< llen; ++k) printf("%d ", resorted[lhs2[k]]);
		printf("=> ");
		if (header)
			for (int k=0; k< rlen; ++k) printf("%s ", attribute_names[resorted[rhs2[k]]].c_str());
		else
			for (int k=0; k< rlen; ++k) printf("%d ", resorted[rhs2[k]]);
		printf("(%f, %f)\n", entropy, mu);
	}

	if (sort_output) {
		delete [] lhs2;
		delete [] rhs2;
	}
	return;
}


void Miner::print_rule(const int *lhs, int llen, int rhs, double entropy, double mu) const
{
	if (!(out||to_std_out)) return;

	int *lhs2;
	if (sort_output) {
		lhs2 = new int[llen];
		memcpy(lhs2, lhs, sizeof(int)*llen);
		sort(lhs2, lhs2+llen);
	}
	else
		lhs2 = (int*)lhs;

	if (out) {
		if (header)
			for (int k=0; k< llen; ++k) fprintf(out, "%s ", attribute_names[resorted[lhs2[k]]].c_str());
		else
			for (int k=0; k< llen; ++k) fprintf(out, "%d ", resorted[lhs2[k]]);
		fprintf(out, "=> ");
		if (header)
			fprintf(out, "%s ", attribute_names[resorted[rhs]].c_str());
		else
			fprintf(out, "%d ", resorted[rhs]);
		fprintf(out, "(%f, %f)\n", entropy, mu);
	}
	else if (to_std_out) {
		if (header)
			for (int k=0; k< llen; ++k) printf("%s ", attribute_names[resorted[lhs2[k]]].c_str());
		else
			for (int k=0; k< llen; ++k) printf("%d ", resorted[lhs2[k]]);
		printf("=> ");
		if (header)
			printf("%s ", attribute_names[resorted[rhs]].c_str());
		else
			printf("%d ", resorted[rhs]);
		printf("(%f, %f)\n", entropy, mu);
	}

	if (sort_output)
		delete [] lhs2;

	return;
}
