/*
 * Template Trie structure
 * Author: Michael Mampaey
 */

#include "intmap.h"
#include <algorithm>
#include <string.h>


using namespace std;



template <class T> class Trie;


template <class T> class Node
{
public:
	Node(T val);
	~Node();

private:
	T val;
	IntMap<Node<T>* > children;
	friend class Trie<T>;
};



template <class T> class Trie
{
public:
	Trie(T defaultval = T());
	~Trie();
	void construct();

	int size() const { return trie_size; }
	Node<T> *add(const int *key, int keylen, T value, Node<T> *start=0);
	T find(int key, Node<T> *start=0) const;
	T find(const int *key, int keylen, int exclude=-1, Node<T> *start=0) const;
	T find(const int *key, int keylen, const int *exclude, int exlen, Node<T> *start=0) const;
	Node<T>* locate(int key, Node<T> *start=0) const;
	Node<T>* locate_rel(int key, Node<T> *start) const;
	pair<int,Node<T>*> locate_rel_hint(int key, Node<T> *start, int hint) const;
	Node<T>* locate(const int *key, int keylen, int exclude=-1, Node<T> *start=0) const;
	Node<T>* locate(const int *key, int keylen, const int *exclude, int exlen, Node<T> *start=0) const;
	bool delete_leaf(const int *key, int keylen, Node<T> *start=0);

	void get_all_values(const int *key, int keylen, void *&all_array, Node<T> *start=0, unsigned loc=0, int l=0);

private:
	void sort_key(int *&key, int keylen) const;
	void unsort_key(int *&key) const;

	int trie_size;
	Node<T> *root;

	const T defaultval;
	T notfound;

	static const bool sort_keys;
	static int *tmp_sort_key;
};

template <class T> const bool Trie<T>::sort_keys = false;
template <class T> int *Trie<T>::tmp_sort_key = 0;




template <class T> inline Node<T>::Node(T val): val(val)
{

}


template <class T> inline Node<T>::~Node()
{
	for (int i=0; i<children.size; ++i) {
		delete children.map[i].second;
	}
}



template <class T> Trie<T>::Trie(T defaultval) : trie_size(0), defaultval(defaultval)
{
	construct();
}


template <class T> inline void Trie<T>::construct() {
	notfound = -1;
	root = new Node<T>(defaultval);
}


template <> inline void Trie<pair<int,double> >::construct() {
	notfound = pair<int,double>(-1,-1);
	root = new Node<pair<int,double> >(defaultval);
}


template <class T> Trie<T>::~Trie()
{
	delete root;
}


template <class T> T Trie<T>::find(const int *key, int keylen, int exclude, Node<T> *start) const
{
	Node<T> *loc = locate(key, keylen, exclude, start);
	if (loc)
		return loc->val;
	else
		return notfound;
}


template <class T> T Trie<T>::find(const int *key, int keylen, const int *exclude, int exlen, Node<T> *start) const
{
	Node<T> *loc = locate(key, keylen, exclude, exlen, start);
	if (loc)
		return loc->val;
	else
		return notfound;
}


template <class T> inline void Trie<T>::sort_key(int *&key, int keylen) const
{
	if (sort_keys) {
		int *sortedkey = new int[keylen];
		memcpy(sortedkey, key, keylen*sizeof(int));
		sort(sortedkey, sortedkey+keylen);
		tmp_sort_key = key;
		key = sortedkey;
	}
}


template <class T> inline void Trie<T>::unsort_key(int *&key) const
{
	if (sort_keys) {
		delete [] key;
		key = tmp_sort_key;
	}
}


template <class T> inline Node<T> *Trie<T>::locate(const int *key, int keylen, int exclude, Node<T>* start) const
{
	int *thekey = (int*)key;
	Node<T> *current;
	if (start)
		current = start;
	else
		current = root;
	if (keylen==0)
		return current;

	int ex = -1;
	if (exclude!=-1) ex = thekey[exclude];

	sort_key(thekey, keylen);

	Node<T>* pit;
	int *sk = thekey;
	thekey += keylen;
	if (exclude==-1) {
		while (sk!=thekey) {
			pit = current->children.get(*sk);
			if (pit==0) {
				unsort_key(thekey);
				return 0;
			}
			else
				current = pit;
			++sk;
		}
	}
	else {
		while (sk!=thekey) {
			if (*sk==ex) {++sk; continue;}
			pit = current->children.get(*sk);
			if (pit==0) {
				unsort_key(thekey);
				return 0;
			}
			else
				current = pit;
			++sk;
		}
	}
	thekey -= keylen;
	unsort_key(thekey);

	return current;
}


template <class T> inline Node<T> *Trie<T>::locate(const int *key, int keylen, const int *exclude, int exlen, Node<T>* start) const
{
	int *thekey = (int*)key;
	Node<T> *current;
	if (start)
		current = start;
	else
		current = root;
	if (keylen==0)
		return current;

		sort_key(thekey, keylen);

	Node<T>* pit;
	int *sk = thekey;
	thekey += keylen;
	if (exclude==0 || exlen==0) {
		while (sk!=thekey) {
			pit = current->children.get(*sk);
			if (pit==0) {
				unsort_key(thekey);
				return 0;
			}
			else
				current = pit;
			++sk;
		}
	}
	else {
		int ex = 0;
		while (sk!=thekey) {
			if (ex!=-1 && *sk==exclude[ex]) {
				++sk;
				++ex;
				if (ex==exlen)
					ex =- 1;
				continue;
			}
			pit = current->children.get(*sk);
			if (pit==0) {
				unsort_key(thekey);
				return 0;
			}
			else
				current = pit;
			++sk;
		}
	}
	thekey -= keylen;
	unsort_key(thekey);

	return current;
}


template <class T> Node<T>* Trie<T>::add(const int *key, int keylen, T value, Node<T> *start)
{
	int *thekey = (int*) key;
	sort_key(thekey, keylen);

	Node<T> *current = start;
	if (current==0)
		current = root;

	Node<T>* pit;
	int *sk = thekey;
	thekey += keylen;
	while (sk!=thekey) {
		pit = current->children.get(*sk);
		if (pit==0) {
			Node<T> *newnode = new Node<T>(defaultval);
			current->children.insert(*sk, newnode);
			current = newnode;
		}
		else
			current = pit;
		++sk;
	}

	current->val = value;
	++trie_size;

	thekey -= keylen;
	unsort_key(thekey);

	return current;
}


template <class T> inline Node<T>* Trie<T>::locate(int key, Node<T> *start) const
{
	Node<T> *current;
	if (start)
		current = start;
	else
		current = root;

	return current->children.get(key);
}


template <class T> inline Node<T>* Trie<T>::locate_rel(int key, Node<T> *start) const
{
	return start->children.get(key);
}


template <class T> inline pair<int, Node<T>*> Trie<T>::locate_rel_hint(int key, Node<T> *start, int hint) const
{
	return start->children.get_r(key, hint);
}


template <class T> T Trie<T>::find(int key, Node<T> *start) const
{
	Node<T> *n = locate(key, start);
	if (n==0)
		return notfound;
	else
		return n->val;
}



template <class T> bool Trie<T>::delete_leaf(const int *key, int keylen, Node<T> *start)
{
	if (keylen==0 && start==0)
		return false;

	int *thekey = (int*)key;

	sort_key(thekey, keylen);

	Node<T> *pit;
	Node<T> *current = start;
	if (current==0)
		current = root;
	Node<T> *prev = 0;
	int *sk = thekey;
	thekey += keylen;
	while (sk!=thekey) {
		pit = current->children.get(*sk);
		if (pit==0) {
			unsort_key(thekey);
			return false;
		}
		else {
			prev = current;
			current = pit;
		}
		++sk;
	}
	thekey -= keylen;
	unsort_key(thekey);

	if (current->children.size==0) { // if leaf
		prev->children.erase(thekey[keylen-1]);
		delete current;
		--trie_size;
		return true;
	}
	else
		return false;
}


template <class T> inline void Trie<T>::get_all_values(const int *key, int keylen, void *&allarray, Node<T> *start, unsigned loc, int l)
{
	static int *thekey = 0;
	T *&all_array = (T*&)(allarray);

	if (l==keylen) {
		all_array[loc] = start->val;
		return;
	}

	if (l==0) {
		thekey = (int*)key;
		sort_key(thekey, keylen);
		start = root;
	}

	if (start==0) {
		if (l==0) {
			unsort_key(thekey);
			thekey = 0;
		}
		all_array = 0;
		return;
	}

	all_array[loc] = start->val;

	for (int k=l; k<keylen; ++k) {
		Node<T> *rooti = locate(thekey[k], start);
		if (rooti==0) {
			all_array = 0;
			if (l==0) {
				unsort_key(thekey);
				thekey = 0;
			}
			return;
		}
		get_all_values(key, keylen, (void*&)all_array, rooti, loc + (1<<k), k + 1);
		if (all_array==0) {
			if (l==0) {
				unsort_key(thekey);
				thekey = 0;
			}
			return;
		}
	}

	if (l==0) {
		unsort_key(thekey);
		thekey = 0;
	}

	return;
}


template <> inline void Trie<pair<int,double> >::get_all_values(const int *key, int keylen, void *&all_array, Node<pair<int,double> > *start, unsigned loc, int l)
{
	static int *thekey = 0;

	int *&full_array = (int*&)(all_array);

	if (l==keylen) {
		full_array[loc] = start->val.first;
		return;
	}

	if (l==0) {
		thekey = (int*)key;
		sort_key(thekey, keylen);
		start = root;
	}

	if (start==0) {
		if (l==0) {
			unsort_key(thekey);
			thekey = 0;
		}
		all_array = 0;
		return;
	}

	full_array[loc] = start->val.first;

	int hint = start->children.size;
	for (int k=l; k<keylen; ++k) {
		pair<int, Node<pair<int,double> > *> rooti = locate_rel_hint(thekey[k], start, hint);
		hint = rooti.first;
		if (rooti.second==0) {
			all_array = 0;
			if (l==0) {
				unsort_key(thekey);
				thekey = 0;
			}
			return;
		}
		get_all_values(key, keylen, all_array, rooti.second, loc + (1<<k), k + 1);
		if (full_array==0) {
			if (l==0) {
				unsort_key(thekey);
				thekey = 0;
			}
			return;
		}
	}

	if (l==0) {
		unsort_key(thekey);
		thekey = 0;
	}

	return;
}
